from .. import settings

from oddrn import Generator

airflow_generator = Generator(data_source=settings.AIRFLOW_SOURCE, cloud=settings.CLOUD)


def get_transformer_oddrn(dag_id: str, task_id: str) -> str:
    return airflow_generator.get_task(dag_id, task_id)

def get_transformer_run_oddrn(dag_id: str, task_id: str, run_id: str) -> str:
    return airflow_generator.get_task_run(dag_id, task_id, run_id)



import os
import sqlparse
from sqlparse.sql import T, TokenList, Parenthesis, Identifier, IdentifierList
from sqlparse.tokens import Punctuation

from . import settings
from .extractors import PostgresExtractor

def _is_in_table(token):
    return _match_on(token, [
        'FROM',
        'INNER JOIN',
        'JOIN',
        'FULL JOIN',
        'FULL OUTER JOIN',
        'LEFT JOIN',
        'LEFT OUTER JOIN',
        'RIGHT JOIN',
        'RIGHT OUTER JOIN'
    ])

def _is_out_table(token):
    return _match_on(token, ['INTO'])

def _match_on(token, keywords):
    return token.match(T.Keyword, values=keywords)

class Parser:
    def __init__(self, task):
        self.extractor = PostgresExtractor(task)
        self._sql = task.sql
        self._inputs = set()
        self._outputs = set()
        self._idx = None
        self.ctes = set()

    def _get_tables(self, tokens):
        def parse_ident(ident: Identifier) -> str:
            token_list = ident.flatten()
            table_name = next(token_list).value
            try:
                dot = next(token_list)
                if dot.match(Punctuation, '.'):
                    table_name += dot.value
                    table_name += next(token_list).value
                else:
                    table_name = f'{table_name}'
            except StopIteration:
                table_name = f'{table_name}'
            return table_name

        self._idx, token = tokens.token_next(idx=self._idx)
        tables = []
        if isinstance(token, IdentifierList):
            gidx = 0
            tables.append(parse_ident(token.token_first(skip_ws=True, skip_cm=True)))
            gidx, punc = token.token_next(gidx, skip_ws=True, skip_cm=True)
            while punc and punc.value == ',':
                gidx, name = token.token_next(gidx, skip_ws=True, skip_cm=True)
                tables.append(parse_ident(name))
                gidx, punc = token.token_next(gidx)
        else:
            tables.append(parse_ident(token))

        return tables

    def is_cte(self, token: T):
        return token.match(T.Keyword.CTE, values=['WITH'])

    def get_response(self):
        self.get_tokens()
        return self.get_inputs(), self.get_outputs()

    def get_tokens(self):
        stats = sqlparse.parse(self._sql)
        tokens = TokenList(stats[0].tokens)
        self._idx, token = tokens.token_next_by(t=T.Keyword)
        while token:
            if self.is_cte(token):
                cte_name, cte_intables = self.parse_cte(tokens)
                for intable in cte_intables:
                    if intable not in self.ctes:
                        self._inputs.add(intable)
            elif _is_in_table(token):
                extracted_tables = self._get_tables(tokens)
                for table in extracted_tables:
                    if table not in self.ctes:
                        self._inputs.add(table)
            elif _is_out_table(token):
                extracted_tables = self._get_tables(tokens)
                self._outputs.add(extracted_tables[0])
            self._idx, token = tokens.token_next_by(t=T.Keyword, idx=self._idx)

    def parse_cte(self, tokens: TokenList):
        gidx, group = tokens.token_next(self._idx, skip_ws=True, skip_cm=True)

        # handle recursive keyword
        if group.match(T.Keyword, values=['RECURSIVE']):
            gidx, group = tokens.token_next(gidx, skip_ws=True, skip_cm=True)

        if not group.is_group:
            return [], None

        # get CTE name
        offset = 1
        cte_name = group.token_first(skip_ws=True, skip_cm=True)
        self.ctes.add(cte_name.value)

        # AS keyword
        offset, as_keyword = group.token_next(offset, skip_ws=True, skip_cm=True)
        if not as_keyword.match(T.Keyword, values=['AS']):
            raise RuntimeError(f"CTE does not have AS keyword at index {gidx}")

        offset, parens = group.token_next(offset, skip_ws=True, skip_cm=True)
        if isinstance(parens, Parenthesis) or parens.is_group:
            # Parse CTE using recursion.
            return cte_name.value, self.recurse(TokenList(parens.tokens)).in_tables
        raise RuntimeError(f"Parens {parens} are not Parenthesis at index {gidx}")

    def get_inputs(self):
        return self.extractor.get_oddrn_list(self._inputs)

    def get_outputs(self):
        return self.extractor.get_oddrn_list(self._outputs)

def get_transformer_name(dag_id: str, task_id: str) -> str:
    return f"{dag_id}.{task_id}"

def get_transformer_run_name(run_id: str, task_id: str) -> str:
    return f"{run_id}.{task_id}"


from .. import settings

from oddrn import Generator

generator = Generator(data_source=settings.POSTGRES_SOURCE, cloud=settings.CLOUD)

def get_table_oddrn(db:str, schema: str, table: str) -> str:
    return generator.get_table(db, schema,  table)



import os

AIRFLOW_SOURCE = "airflow"
POSTGRES_SOURCE = "postgresql"

CATALOG_URL = os.getenv("CATALOG_URL", None)

CLOUD_TYPE = os.getenv("CLOUD_TYPE", "aws")
CLOUD_REGION = os.getenv("CLOUD_REGION", "region_id")
CLOUD_ACCOUNT = os.getenv("CLOUD_ACCOUNT", "account_id")

CLOUD = {
    "type": CLOUD_TYPE,
    "region": CLOUD_REGION,
    "account": CLOUD_ACCOUNT
}


from ..oddrn.postgres import get_table_oddrn

from .. import settings

class PostgresExtractor:
    def __init__(self, task):
        self._db = task.postgres_conn_id

    def get_oddrn_list(self, tables):
        response = []
        for table in tables:
            source = table.split(".")
            if len(source) > 1:
                oddrn = get_table_oddrn(self._db, source[0],  source[1])
            else:
                oddrn = get_table_oddrn(self._db, "public", source[0])
            response.append(oddrn)
        return response


import requests
import sqlparse
from sqlparse.sql import T, TokenList, Parenthesis, Identifier, IdentifierList
from sqlparse.tokens import Punctuation
from pkg_resources import parse_version

import airflow.models
from airflow.version import version as AIRFLOW_VERSION

from .settings import CATALOG_URL
from .utils import *
from .oddrn.airflow import get_transformer_oddrn, get_transformer_run_oddrn

# Corrects path of import for Airflow different versions
if parse_version(AIRFLOW_VERSION) >= parse_version("1.10.11"):
    from airflow import LoggingMixin
else:
    from airflow.utils.log.logging_mixin import LoggingMixin


class DAG(airflow.models.DAG, LoggingMixin):
    def create_dagrun(self, *args, **kwargs):
        dagrun = super(DAG, self).create_dagrun(*args, **kwargs)
        data = []
        for task_id, task in self.task_dict.items():
            if hasattr(task, "sql"):
                data.extend([self._add_run(task, dagrun), self._add_transformer(task)])
        self._send_data(data)
        return dagrun

    def handle_callback(self, *args, **kwargs):
        try:
            dagrun = args[0]
            transformer_runs = []
            for task_id, task in self.task_dict.items():
                transformer_runs.append(self._add_run(task, dagrun))
            self._send_data(transformer_runs)
        except Exception as e:
            self.log.error(
                f'Failed to record dagrun callback: {e} '
                f'dag_id={self.dag_id}',
                exc_info=True
            )

        return super().handle_callback(*args)

    def _send_data(self, data: dict):
        url = self.default_args.get("data_catalog_host", CATALOG_URL)
        r = requests.post(url, json=data)
        if r.status_code == 200:
            self.log.info(f"Data transfer success")
        else:
            self.log.error(f"Error on catalog request. Code: {r.status_code}, Message: {r.text}")

    def _add_transformer(self, task) -> dict:
        sql_parser = Parser(task)
        dag = task.dag
        inputs, outputs = sql_parser.get_response()
        data = {
            "oddrn": get_transformer_oddrn(dag.dag_id, task.task_id),
            "name": get_transformer_name(dag.dag_id, task.task_id),
            "description": task.doc_md if task.doc_md else dag.description,
            "owner": dag.owner,
            "type": "JOB",
            "data_transformer": {
                "source_code_url": "source",
                "sql": task.sql,
                "inputs": inputs,
                "outputs": outputs
            },
            "metadata": {
                "catchup": dag.cathup
            }
        }
        return data

    def _add_run(self, task, dagrun) -> dict:
        dag = dagrun.dag
        data = {
            "oddrn": get_transformer_run_oddrn(dag.dag_id, task.task_id, dagrun.run_id),
            "name": get_transformer_run_name(dagrun.run_id, task.task_id),
            "description": task.doc_md if task.doc_md else dag.description,
            "owner": dag.owner,
            "type": "JOB_RUN",
            "data_transformer_run": {
                "transformer_oddrn": get_transformer_oddrn(dag.dag_id, task.task_id),
                "start_time": dagrun.start_date.isoformat(),
                "end_time": dagrun.end_date.isoformat(),
                "status": "OTHER", # map status,
                "status_reason": "" # get status reason?
            },
            "metadata": {
                "catchup": dag.cathup
            }
        }
        return data




from setuptools import find_packages, setup

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = [
    "attrs>=19.3",
    "requests>=2.24.0",
    "sqlparse==0.4.1",
    "oddrn>=0.0.3"
]

extras_require = {}

setup(
    name="odd_airflow",
    version="0.0.7",
    description="ODD adapter to Airflow",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Provectus ODD team",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    extras_require=extras_require,
    python_requires=">=3.6",
    zip_safe=False,
    keywords="odd-airflow",
)